import json
from pathlib import Path

from pydantic import Field, ConfigDict
from pydantic_settings import BaseSettings

from . import utils
from .constants import __app_name__

GLOBAL_CONFIG = Path.home() / f".{__app_name__}"
PLANS_DIR = GLOBAL_CONFIG / "plans"
GLOBAL_SETTINGS_FILE = GLOBAL_CONFIG / "settings.json"


class SystemSettings(BaseSettings):
    """
    Global system settings (not editable by user)
    """

    recipe_library: Path | None = Field(
        default=None,
        description=(
            "Path to the user's recipe library. If this is None, it means "
            "that the user still needs to run `chef init`."
        ),
    )

    @property
    def user_config_dir(self) -> Path:
        """
        Returns:
            A path to the hidden configuration folder in the user's
            recipe library
        """
        if self.recipe_library is None:
            raise NotInitialised

        return self.recipe_library / f".{__app_name__}"

    @property
    def project_settings(self):
        return self.user_config_dir / "settings.json"


class ProjectSettings(BaseSettings):
    # project settings
    merge_ingredients: bool = Field(
        default=True,
        description=(
            "If True, try to squash together similar ingredients when making "
            "a shopping list "
            "e.g. `5, g, garlic` + `10, g, garlic` -> `15, g, garlic`"
        ),
    )
    model_config = ConfigDict(extra="ignore")


class Settings(BaseSettings):
    """
    NOTE TO SELF:

    I _do_ want to have separate global and project settings, because I want to
    be able to check the .yes-chef folder into git.
    """

    system: SystemSettings = SystemSettings()
    project: ProjectSettings = ProjectSettings()

    def save(self) -> None:
        """
        Saves the system and project settings to their respective files.

        Raises:
            NotInitialised: if the recipe library is not configured.
        """
        utils.touch(GLOBAL_SETTINGS_FILE)
        with open(GLOBAL_SETTINGS_FILE, "w") as file:
            file.write(self.system.model_dump_json())

        with open(self.system.project_settings, "w") as file:
            file.write(self.project.model_dump_json())

    @classmethod
    def load(cls) -> "Settings":
        """
        Load the system settings and the project settings. Should always
        succeed, even if the recipe library is not initialised. Otherwise the
        program will fail on startup and we won't be able to initialise the
        library.

        Returns:
            Settings object
        """
        system_settings = SystemSettings(
            **get_or_create_json(GLOBAL_SETTINGS_FILE)
        )
        try:
            project_settings = ProjectSettings(
                **get_or_create_json(system_settings.project_settings)
            )
        except NotInitialised:
            project_settings = ProjectSettings()
        return cls(
            system=system_settings,
            project=project_settings,
        )


def get_or_create_json(path: Path) -> dict:
    """
    Reads JSON from a file. Creates the file if it doesn't exist.

    Args:
        path: the JSON file to open

    Returns:
        A dictionary of native Python data types

    Raises:
        json.JSONDecodeError: if the JSON is not valid
        ValueError: if the JSON starts with a "[" instead of a "{"
    """
    try:
        with open(path) as file:
            json_str = file.read()
    except FileNotFoundError:
        utils.touch(path)
        with open(path, "w") as file:
            file.write("{}")
        return {}

    try:
        data = json.loads(json_str)
        assert isinstance(data, dict)
    except AssertionError:
        raise ValueError(f"JSON shouldn't be an array: {json_str[:10]}")
    else:
        return data


class NotInitialised(Exception):
    pass



import asyncio
from pathlib import Path

import aiofiles

from ... import api
from ...api import Recipe, RecipeSerializer
from ...api import Settings
from ...api.recipe.serializer import Format
from ...api.recipe.utils import capitalise
from ...api.utils import clean_filename


async def export_recipes():
    settings = Settings.load()
    markdown_folder = settings.system.recipe_library / "md"  # todo: dedupe this
    recipes = await Recipe.load_all()
    print(f"loaded {len(recipes)} recipes")
    await asyncio.gather(
        *(_serialize(recipe, markdown_folder) for recipe in recipes)
    )
    await create_readme(
        {
            _markdown_filename(recipe, markdown_folder): recipe
            for recipe in recipes
        }
    )


async def create_readme(recipes: dict[Path, Recipe]):
    """
    Args:
        recipes: dict of markdown filenames : recipe objects
    """
    settings = Settings.load()

    readme = ["# Recipes"]

    recipes_list = []
    for md_filename in sorted(recipes):
        recipe = recipes[md_filename]
        recipes_list.append(
            f"- [{capitalise(recipe.name)}](md/{md_filename.name}) "
            f"by {recipe.author}"
        )
    recipes_list_str = "\n".join(recipes_list)
    readme.append(recipes_list_str)

    readme.append(
        f"Generated by [`yes-chef`](https://github.com/binnev/yes-chef) "
        f"version {api.__version__}"
    )

    readme_str = "\n\n".join(readme)

    async with aiofiles.open(
        settings.system.recipe_library / "README.md", "w"
    ) as file:
        await file.write(readme_str)


async def _serialize(recipe: Recipe, markdown_dir: Path):
    filename = _markdown_filename(recipe, markdown_dir)
    serialized = RecipeSerializer().serialize(recipe, Format.MARKDOWN)
    async with aiofiles.open(filename, "w") as file:
        await file.write(serialized)


def _markdown_filename(recipe: Recipe, markdown_dir: Path) -> Path:
    return clean_filename(markdown_dir / f"{recipe.name}.md")



"""
Handles the merging of multiple ingredient lists into one shopping list.
"""

from pydantic import BaseModel, Field

from .recipe import Recipe
from .unit import normalise

Number = int | float
Unit = str
IngredientName = str
RecipeName = str


class MergedIngredient(BaseModel):
    """
    Container class to hold merged Ingredients. Handles a mixture of
    amountless, unitless, and fully populated Ingredients. It is not possible
    to merge a fully populated unit with an amountless unit -- e.g. "2, kg,
    potatoes" and "potatoes". This is why we need the following attributes

    Attributes:
        amountless: for Ingredients without an amount, we just want to keep
            track of which recipes we need them for. So that instead of seeing
            "coriander" in the shopping list, we see "coriander: enough for
            recipes A, B, C"
        unitless: for Ingredients with no units, we can simply add the amounts,
            e.g. "2 apples" + "3 apples" = "5 apples"
        units: a mapping of amounts to units. For units that can't be merged
            like "cm" and "g"
    """

    # from "apples" -> display as "enough for x"
    amountless: list[RecipeName] = []

    # from "4 apples"
    unitless: Number = 0

    # from "4 kg apples"
    units: dict[Unit, Number] = Field(default_factory=dict)


ShoppingList = dict[IngredientName, MergedIngredient]


def merge_recipes(recipes: list[Recipe]) -> ShoppingList:
    """
    Create a combined shopping list by merging the ingredients from multiple
    recipes.

    Args:
        recipes: the list of recipes we want to generate a shopping list for

    Returns:
        the merged shopping list
    """
    shopping_list = {}
    for recipe in recipes:
        for ing in recipe.ingredients:
            merged = shopping_list.get(ing.name, MergedIngredient())
            match [ing.amount, ing.unit]:
                case [0, ""]:
                    merged.amountless.append(recipe.name)
                case [amount, ""]:
                    merged.unitless += amount
                case [amount, unit]:
                    amount, unit = normalise(amount, unit)
                    merged.units[unit] = merged.units.get(unit, 0) + amount
            shopping_list[ing.name] = merged

    return shopping_list



import enum
import typing as t

from . import formats
from .recipe import Recipe


class Format(enum.Enum):
    YAML = enum.auto()
    MARKDOWN = enum.auto()


# function which takes a recipe and serializes it to some string format
SerializerFunc = t.Callable[[Recipe], str]


class RecipeSerializer:
    """
    RecipeSerializer.serialize(recipe, fmt=Format.MARKDOWN)
    """

    _handlers: dict[Format, SerializerFunc] = {}

    def serialize(self, recipe: Recipe, fmt: Format) -> str:
        handler = self._get_handler(fmt)
        result = handler(recipe)
        return result

    def _get_handler(self, fmt: Format) -> t.Callable:
        return self._handlers[fmt]

    @classmethod
    def register(cls, fmt: Format, handler: SerializerFunc):
        cls._handlers[fmt] = handler


RecipeSerializer.register(Format.YAML, formats.serialize_yaml)
RecipeSerializer.register(Format.MARKDOWN, formats.serialize_markdown)



import typing as t
from functools import wraps
from typing import TypeVar

import inquirer
import typer

from src.api import Settings

T = TypeVar("T")


def echo(
    s: str,
    nl: bool = True,
    err: bool = False,
    color: t.Optional[bool] = None,
    **styles: t.Any,
):
    typer.secho(
        s.expandtabs(4),
        nl=nl,
        err=err,
        color=color,
        **styles,
    )


def multiple_choice_menu(prompt: str, choices: dict[str, T]) -> T:
    """
    This syntax is pretty cumbersome and nested, so I'm putting it in a
    function.
    :param prompt: The text that will be displayed to the user
    :param choices: dict of {obj_name: obj} (obj_name will be displayed to user)
    :return: the selected object
    """
    key = inquirer.prompt(
        [
            inquirer.List(
                "_",
                message=prompt,
                choices=choices.keys(),
            )
        ]
    )["_"]
    return choices[key]


INIT_MSG = """
You need to initialise your recipe library by running:

    chef init <path-to-your-recipe-library>
"""


def requires_library_init(func):
    """
    Decorator that checks whether the recipe library has been initialised
    before certain functions can be called.
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        settings = Settings.load()
        if settings.system.recipe_library is None:
            typer.secho(INIT_MSG)
            raise typer.Exit(1)
        return func(*args, **kwargs)

    return wrapped



from pathlib import Path

from .settings import Settings
from .constants import __app_name__


def init_library(path: Path):
    settings = Settings.load()
    settings.system.recipe_library = path
    settings.save()

    for p in [
        path,
        path / "yaml",
        path / "md",
        path / f".{__app_name__}",
    ]:
        Path.mkdir(p, parents=True, exist_ok=True)  # ensure the dir exists

    if not settings.system.project_settings.exists():
        with open(settings.system.project_settings, "w") as file:
            file.write("{}")  # empty json



from ..ingredient import parse_ingredient_str


def preprocess_yaml(yaml_dict: dict) -> dict:
    """
    Convert all the ingredient strings into ingredient dicts, so that we can
    pass the result to pydantic models.

    :param yaml_dict: YAML recipe loaded into dict form
    :raises: ParseIngredientError
    """
    if "ingredients" in yaml_dict:
        yaml_dict["ingredients"] = list(
            map(parse_ingredient_str, yaml_dict["ingredients"])
        )
    return yaml_dict


def capitalise(s: str) -> str:
    """
    Just capitalise the first character; don't do anything else
    """
    if not s:
        return s

    return s[0].capitalize() + s[1:]



"""
Handles normalising and converting between units.
"""

Number = int | float


def normalise(amount: Number, unit: str) -> tuple[Number, str]:
    """
    Converts an amount + unit into the preferred internal format (grams or
    millilitres) if possible

    Handles:
        - alias: (pounds -> lb)
        - system (imperial -> metric): (lb -> kg)
        - denomination: (5kg -> 5000g)

    Args:
        amount: the input amount e.g. 2.5
        unit: the input unit e.g. "kilos"

    Returns:
        The converted amount and unit
    """
    unit = unit.lower()
    unit = SINGULARS.get(unit, unit)
    unit = PREFERRED_ALIASES.get(unit, unit)
    multiplier, internal_unit = CONVERSIONS.get(unit, (1, unit))
    return amount * multiplier, internal_unit


# These are the units we will use internally. I've chosen the smallest metric
# denomination so that we can use integers.
GRAM = "g"
ML = "ml"
TSP = "tsp"
INTERNAL_UNITS = {GRAM, ML, TSP}

# The preferred alias is the key, the alternatives are the value
ALIASES = {
    GRAM: ["gram", "grams", "gs", "gr"],
    ML: [
        "millilitre",
        "millilitres",
        "milliliter",
        "milliliters",
        "mls",
    ],
    "kg": ["kilo", "kilos", "kgs", "kilogram", "kilograms"],
    "l": ["litre", "litres", "liter", "liters", "L"],
    "lb": ["pound", "pounds", "lbs"],
    "oz": ["ounce", "ounces"],
    "cup": ["cups"],
    "fl oz": ["fluid ounce", "fluid ounces"],
    TSP: ["teaspoon", "teaspoons", "tsps"],
    "tbsp": ["tbsps", "tablespoon", "tablespoons"],
}
PREFERRED_ALIASES = {
    alternative: preferred
    for preferred, alternatives in ALIASES.items()
    for alternative in alternatives
}

# Here we relate every unit to our base units of weight and volume (g, ml)
CONVERSIONS = {
    GRAM: (1, GRAM),
    ML: (1, ML),
    TSP: (1, TSP),
    "kg": (1000, GRAM),
    "l": (1000, ML),
    "oz": (28, GRAM),
    "lb": (454, GRAM),
    "cup": (240, ML),
    "fl oz": (30, ML),
    "tbsp": (3, TSP),
}

# singular to the left; plural to the right
PLURALS = {
    "clove": "cloves",  # of garlic
    "sprig": "sprigs",
}
SINGULARS = {plural: singular for singular, plural in PLURALS.items()}



"""
new.py
"""

import typer

from .utils import requires_library_init
from .routines.recipe_wizard import recipe_wizard
from .. import api
from ..api import RecipeSerializer, Settings
from ..api.recipe.serializer import Format
from ..api.utils import clean_filename

app = typer.Typer()


@app.command(name="plan")
@requires_library_init
def new_plan():
    """
    Create a new plan
    """
    api.Plan.new()
    typer.secho("created new plan\n")


@app.command(name="recipe")
@requires_library_init
def new_recipe():
    """
    New recipe wizard
    """
    recipe = recipe_wizard()
    yaml_str = RecipeSerializer().serialize(recipe, Format.YAML)
    settings = Settings.load()
    filename = clean_filename(
        settings.system.recipe_library / "yaml" / f"{recipe.name}.yaml"
    )
    with open(filename, "w") as file:
        file.write(yaml_str)

    print(f"Created recipe {filename}")
